@@ -8,8 +8,7 @@ import uuid
 from future.standard_library import install_aliases
 
 from pyngrok import process, conf
-from pyngrok.conf import PyngrokConfig
-from pyngrok.exception import PyngrokNgrokHTTPError, PyngrokNgrokURLError, PyngrokSecurityError
+from pyngrok.exception import PyngrokNgrokHTTPError, PyngrokNgrokURLError, PyngrokSecurityError, PyngrokError
 from pyngrok.installer import install_ngrok, install_default_config
 
 install_aliases()
@@ -27,39 +26,47 @@ except ImportError:  # pragma: no cover
 
 __author__ = "Alex Laird"
 __copyright__ = "Copyright 2020, Alex Laird"
-__version__ = "4.1.6"
+__version__ = "4.1.15"
 
 logger = logging.getLogger(__name__)
 
 
 class NgrokTunnel:
     """
-    An object containing information about a :code:`ngrok` tunnel.
+    An object containing information about a ``ngrok`` tunnel.
 
     :var name: The name of the tunnel.
     :vartype name: str
     :var proto: A valid `tunnel protocol <https://ngrok.com/docs#tunnel-definitions>`_.
     :vartype proto: str
-    :var uri: The tunnel URI, a relative path that can be used to make requests to the :code:`ngrok` web interface.
+    :var uri: The tunnel URI, a relative path that can be used to make requests to the ``ngrok`` web interface.
     :vartype uri: str
-    :var public_url: The public :code:`ngrok` URL.
+    :var public_url: The public ``ngrok`` URL.
     :vartype public_url: str
     :var config: The config for the tunnel.
     :vartype config: dict
     :var metrics: Metrics for `the tunnel <https://ngrok.com/docs#list-tunnels>`_.
     :vartype metrics: dict
+    :var pyngrok_config: The ``pyngrok`` configuration to use with ``ngrok``.
+    :vartype pyngrok_config: PyngrokConfig
+    :var api_url: The API URL for the ``ngrok`` web interface.
+    :vartype api_url: str
     """
 
-    def __init__(self, data=None):
+    def __init__(self, data=None, pyngrok_config=None, api_url=None):
         if data is None:
             data = {}
+        if pyngrok_config is None:
+            pyngrok_config = conf.DEFAULT_PYNGROK_CONFIG
 
-        self.name = data["name"] if data else None
-        self.proto = data["proto"] if data else None
-        self.uri = data["uri"] if data else None
-        self.public_url = data["public_url"] if data else None
-        self.config = data["config"] if data else {}
-        self.metrics = data["metrics"] if data else None
+        self.name = data.get("name")
+        self.proto = data.get("proto")
+        self.uri = data.get("uri")
+        self.public_url = data.get("public_url")
+        self.config = data.get("config", {})
+        self.metrics = data.get("metrics", {})
+        self.pyngrok_config = pyngrok_config
+        self.api_url = api_url
 
     def __repr__(self):
         return "<NgrokTunnel: \"{}\" -> \"{}\">".format(self.public_url, self.config["addr"]) if self.config.get(
@@ -69,13 +76,28 @@ class NgrokTunnel:
         return "NgrokTunnel: \"{}\" -> \"{}\"".format(self.public_url, self.config["addr"]) if self.config.get(
             "addr", None) else "<pending Tunnel>"
 
+    def refresh_metrics(self):
+        """
+        Refresh the metrics from the tunnel.
+        """
+        if self.api_url is None:
+            raise PyngrokError("\"api_url\" was not initialized with this NgrokTunnel, so this method cannot be used.")
+
+        data = api_request("{}{}".format(self.api_url, self.uri), method="GET", data=None,
+                           timeout=self.pyngrok_config.request_timeout)
+
+        if "metrics" not in data:
+            raise PyngrokError("The ngrok API did not return \"metrics\" in the response")
+
+        self.metrics = data["metrics"]
+
 
 def ensure_ngrok_installed(ngrok_path):
     """
-    Ensure :code:`ngrok` is installed at the given path, downloading and installing the binary for
+    Ensure ``ngrok`` is installed at the given path, downloading and installing the binary for
     the current system if not.
 
-    :param ngrok_path: The path to the :code:`ngrok` binary.
+    :param ngrok_path: The path to the ``ngrok`` binary.
     :type ngrok_path: str
     """
     if not os.path.exists(ngrok_path):
@@ -87,19 +109,21 @@ def ensure_ngrok_installed(ngrok_path):
 
 def set_auth_token(token, pyngrok_config=None):
     """
-    Set the :code:`ngrok` auth token in the config file, enabling authenticated features (for instance,
+    Set the ``ngrok`` auth token in the config file, enabling authenticated features (for instance,
     more concurrent tunnels, custom subdomains, etc.).
 
-    If :code:`ngrok` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s :code:`ngrok_path`, calling this method
-    will first download and install :code:`ngrok`.
+    If ``ngrok`` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s ``ngrok_path``, calling this method
+    will first download and install ``ngrok``.
 
     :param token: The auth token to set.
     :type token: str
-    :param pyngrok_config: The :code:`pyngrok` configuration to use when interacting with the :code:`ngrok` binary.
+    :param pyngrok_config: The ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
+        defaults to ``conf.DEFAULT_PYNGROK_CONFIG`` (which can be overridden instead,
+        `as shown here <index.html#config-file>`_).
     :type pyngrok_config: PyngrokConfig, optional
     """
     if pyngrok_config is None:
-        pyngrok_config = PyngrokConfig()
+        pyngrok_config = conf.DEFAULT_PYNGROK_CONFIG
 
     ensure_ngrok_installed(pyngrok_config.ngrok_path)
 
@@ -108,21 +132,23 @@ def set_auth_token(token, pyngrok_config=None):
 
 def get_ngrok_process(pyngrok_config=None):
     """
-    Retrieve the current :code:`ngrok` process for the given config's :code:`ngrok_path`.
+    Retrieve the current ``ngrok`` process for the given config's ``ngrok_path``.
 
-    If :code:`ngrok` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s :code:`ngrok_path`, calling this method
-    will first download and install :code:`ngrok`.
+    If ``ngrok`` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s ``ngrok_path``, calling this method
+    will first download and install ``ngrok``.
 
-    If :code:`ngrok` is not running, calling this method will first start a process with
+    If ``ngrok`` is not running, calling this method will first start a process with
     :class:`~pyngrok.conf.PyngrokConfig`.
 
-    :param pyngrok_config: The :code:`pyngrok` configuration to use when interacting with the :code:`ngrok` binary.
+    :param pyngrok_config: The ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
+        defaults to ``conf.DEFAULT_PYNGROK_CONFIG`` (which can be overridden instead,
+        `as shown here <index.html#config-file>`_).
     :type pyngrok_config: PyngrokConfig, optional
-    :return: The :code:`ngrok` process.
+    :return: The ``ngrok`` process.
     :rtype: NgrokProcess
     """
     if pyngrok_config is None:
-        pyngrok_config = PyngrokConfig()
+        pyngrok_config = conf.DEFAULT_PYNGROK_CONFIG
 
     ensure_ngrok_installed(pyngrok_config.ngrok_path)
 
@@ -131,13 +157,13 @@ def get_ngrok_process(pyngrok_config=None):
 
 def connect(port="80", proto="http", name=None, options=None, pyngrok_config=None):
     """
-    Establish a new :code:`ngrok` tunnel to the given port and protocol, returning the connected
+    Establish a new ``ngrok`` tunnel for the given protocol to the given port, returning the connected
     public URL that tunnels to the local port.
 
-    If :code:`ngrok` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s :code:`ngrok_path`, calling this method
-    will first download and install :code:`ngrok`.
+    If ``ngrok`` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s ``ngrok_path``, calling this method
+    will first download and install ``ngrok``.
 
-    If :code:`ngrok` is not running, calling this method will first start a process with
+    If ``ngrok`` is not running, calling this method will first start a process with
     :class:`~pyngrok.conf.PyngrokConfig`.
 
     :param port: The local port to which the tunnel will forward traffic, defaults to "80". Can also be
@@ -150,7 +176,9 @@ def connect(port="80", proto="http", name=None, options=None, pyngrok_config=Non
     :param options: Parameters passed to `configuration for the ngrok
         tunnel <https://ngrok.com/docs#tunnel-definitions>`_.
     :type options: dict[str, str], optional
-    :param pyngrok_config: The :code:`pyngrok` configuration to use when interacting with the :code:`ngrok` binary.
+    :param pyngrok_config: The ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
+        defaults to ``conf.DEFAULT_PYNGROK_CONFIG`` (which can be overridden instead,
+        `as shown here <index.html#config-file>`_).
     :type pyngrok_config: PyngrokConfig, optional
     :return: The connected public URL.
     :rtype: str
@@ -158,7 +186,7 @@ def connect(port="80", proto="http", name=None, options=None, pyngrok_config=Non
     if options is None:
         options = {}
     if pyngrok_config is None:
-        pyngrok_config = PyngrokConfig()
+        pyngrok_config = conf.DEFAULT_PYNGROK_CONFIG
 
     port = str(port)
     if not name:
@@ -178,10 +206,11 @@ def connect(port="80", proto="http", name=None, options=None, pyngrok_config=Non
 
     logger.debug("Connecting tunnel with options: {}".format(options))
 
-    tunnel = NgrokTunnel(api_request("{}/api/{}".format(api_url, "tunnels"), method="POST", data=options,
-                                     timeout=pyngrok_config.request_timeout))
+    tunnel = NgrokTunnel(api_request("{}/api/tunnels".format(api_url), method="POST", data=options,
+                                     timeout=pyngrok_config.request_timeout),
+                         pyngrok_config=pyngrok_config, api_url=api_url)
 
-    if proto == "http" and ("bind_tls" not in options or options["bind_tls"] is not False):
+    if proto == "http" and not options.get("bind_tls", False):
         tunnel.public_url = tunnel.public_url.replace("https", "http")
 
     return tunnel.public_url
@@ -189,21 +218,23 @@ def connect(port="80", proto="http", name=None, options=None, pyngrok_config=Non
 
 def disconnect(public_url, pyngrok_config=None):
     """
-    Disconnect the :code:`ngrok` tunnel for the given URL.
+    Disconnect the ``ngrok`` tunnel for the given URL.
 
-    If :code:`ngrok` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s :code:`ngrok_path`, calling this method
-    will first download and install :code:`ngrok`.
+    If ``ngrok`` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s ``ngrok_path``, calling this method
+    will first download and install ``ngrok``.
 
-    If :code:`ngrok` is not running, calling this method will first start a process with
+    If ``ngrok`` is not running, calling this method will first start a process with
     :class:`~pyngrok.conf.PyngrokConfig`.
 
     :param public_url: The public URL of the tunnel to disconnect.
     :type public_url: str
-    :param pyngrok_config: The :code:`pyngrok` configuration to use when interacting with the :code:`ngrok` binary.
+    :param pyngrok_config: The ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
+        defaults to ``conf.DEFAULT_PYNGROK_CONFIG`` (which can be overridden instead,
+        `as shown here <index.html#config-file>`_).
     :type pyngrok_config: PyngrokConfig, optional
     """
     if pyngrok_config is None:
-        pyngrok_config = PyngrokConfig()
+        pyngrok_config = conf.DEFAULT_PYNGROK_CONFIG
 
     api_url = get_ngrok_process(pyngrok_config).api_url
 
@@ -220,49 +251,73 @@ def disconnect(public_url, pyngrok_config=None):
 
 def get_tunnels(pyngrok_config=None):
     """
-    Retrieve a list of active :code:`ngrok` tunnels for the given config's :code:`ngrok_path`.
+    Retrieve a list of active ``ngrok`` tunnels for the given config's ``ngrok_path``.
 
-    If :code:`ngrok` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s :code:`ngrok_path`, calling this method
-    will first download and install :code:`ngrok`.
+    If ``ngrok`` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s ``ngrok_path``, calling this method
+    will first download and install ``ngrok``.
 
-    If :code:`ngrok` is not running, calling this method will first start a process with
+    If ``ngrok`` is not running, calling this method will first start a process with
     :class:`~pyngrok.conf.PyngrokConfig`.
 
-    :param pyngrok_config: The :code:`pyngrok` configuration to use when interacting with the :code:`ngrok` binary.
+    :param pyngrok_config: The ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
+        defaults to ``conf.DEFAULT_PYNGROK_CONFIG`` (which can be overridden instead,
+        `as shown here <index.html#config-file>`_).
     :type pyngrok_config: PyngrokConfig, optional
-    :return: The active :code:`ngrok` tunnels.
+    :return: The active ``ngrok`` tunnels.
     :rtype: list[NgrokTunnel]
     """
     if pyngrok_config is None:
-        pyngrok_config = PyngrokConfig()
+        pyngrok_config = conf.DEFAULT_PYNGROK_CONFIG
 
     api_url = get_ngrok_process(pyngrok_config).api_url
 
     tunnels = []
-    for tunnel in api_request("{}/api/{}".format(api_url, "tunnels"), method="GET", data=None,
+    for tunnel in api_request("{}/api/tunnels".format(api_url), method="GET", data=None,
                               timeout=pyngrok_config.request_timeout)["tunnels"]:
-        tunnels.append(NgrokTunnel(tunnel))
+        tunnels.append(NgrokTunnel(tunnel, pyngrok_config=pyngrok_config, api_url=api_url))
 
     return tunnels
 
 
 def kill(pyngrok_config=None):
     """
-    Terminate the :code:`ngrok` processes, if running, for the given config's :code:`ngrok_path`. This method will not
+    Terminate the ``ngrok`` processes, if running, for the given config's ``ngrok_path``. This method will not
     block, it will just issue a kill request.
 
-    :param pyngrok_config: The :code:`pyngrok` configuration to use when interacting with the :code:`ngrok` binary.
+    :param pyngrok_config: The ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
+        defaults to ``conf.DEFAULT_PYNGROK_CONFIG`` (which can be overridden instead,
+        `as shown here <index.html#config-file>`_).
     :type pyngrok_config: PyngrokConfig, optional
     """
     if pyngrok_config is None:
-        pyngrok_config = PyngrokConfig()
+        pyngrok_config = conf.DEFAULT_PYNGROK_CONFIG
 
     process.kill_process(pyngrok_config.ngrok_path)
 
 
 def api_request(url, method="GET", data=None, params=None, timeout=4):
     """
-    Invoke an API request to the given URI, returning JSON data from the response as a dict.
+    Invoke an API request to the given URL, returning JSON data from the response as a dict.
+
+    One use for this method is making requests to ``ngrok`` tunnels:
+
+    .. code-block:: python
+
+        from pyngrok import ngrok
+
+        public_url = ngrok.connect()
+        response = ngrok.api_request("{}/some-route".format(public_url),
+                                     method="POST", data={"foo": "bar"})
+
+    Another is making requests to the ``ngrok`` API itself:
+
+    .. code-block:: python
+
+        from pyngrok import ngrok
+
+        api_url = ngrok.get_ngrok_process().api_url
+        response = ngrok.api_request("{}/api/requests/http".format(api_url),
+                                     params={"tunnel_name": "foo"})
 
     :param url: The request URL.
     :type url: str
@@ -271,7 +326,7 @@ def api_request(url, method="GET", data=None, params=None, timeout=4):
     :param data: The request body.
     :type data: dict, optional
     :param params: The URL parameters.
-    :type params: list[str], optional
+    :type params: dict, optional
     :param timeout: The request timeout, in seconds.
     :type timeout: float, optional
     :return: The response from the request.
@@ -299,7 +354,7 @@ def api_request(url, method="GET", data=None, params=None, timeout=4):
 
         status_code = response.getcode()
         logger.debug("Response status code: {}".format(status_code))
-        logger.debug("Response: {}".format(response_data))
+        logger.debug("Response: {}".format(response_data.strip()))
 
         if str(status_code)[0] != "2":
             raise PyngrokNgrokHTTPError("ngrok client API returned {}: {}".format(status_code, response_data), url,
@@ -315,7 +370,7 @@ def api_request(url, method="GET", data=None, params=None, timeout=4):
 
         status_code = e.getcode()
         logger.debug("Response status code: {}".format(status_code))
-        logger.debug("Response: {}".format(response_data))
+        logger.debug("Response: {}".format(response_data.strip()))
 
         raise PyngrokNgrokHTTPError("ngrok client exception, API returned {}: {}".format(status_code, response_data),
                                     e.url,
@@ -326,9 +381,13 @@ def api_request(url, method="GET", data=None, params=None, timeout=4):
 
 def run(args=None):
     """
-    Start a blocking :code:`ngrok` process with the default binary and the passed args.
+    Ensure ``ngrok`` is installed in the default location, then call :func:`~pyngrok.process.run_process`.
 
-    :param args: Arguments to be passed to the :code:`ngrok` process.
+    This method is meant for interacting with ``ngrok`` from the command line and is not necessarily
+    compatible with non-blocking API methods. For that, use :mod:`~pyngrok.ngrok`'s interface methods (like
+    :func:`~pyngrok.ngrok.connect`), or use :func:`~pyngrok.process.get_process`.
+
+    :param args: Arguments to be passed to the ``ngrok`` process.
     :type args: list[str], optional
     """
     if args is None:
@@ -341,7 +400,12 @@ def run(args=None):
 
 def main():
     """
-    Entry point for the package's :code:`console_scripts`.
+    Entry point for the package's ``console_scripts``. This initializes a call from the command
+    line and invokes :func:`~pyngrok.ngrok.run`.
+
+    This method is meant for interacting with ``ngrok`` from the command line and is not necessarily
+    compatible with non-blocking API methods. For that, use :mod:`~pyngrok.ngrok`'s interface methods (like
+    :func:`~pyngrok.ngrok.connect`), or use :func:`~pyngrok.process.get_process`.
     """
     run(sys.argv[1:])
 
